import unittest
from pyappi.client import use_test_client, RealtimeHttpClient
from pyappi.client.sync import RawClientEvents, UserChanges
import time


class TestSync(unittest.TestCase):
    def setUp(self):
        use_test_client()

        RealtimeHttpClient("test_sync").delete()

    def test_tail(self):
        with UserChanges(interval=.3):

            time.sleep(1)

            with RealtimeHttpClient("test_sync") as doc:
                u = doc.update

                u.nested.value = "CAsfawojgaw"
                u.name = "TEST"
                u.here = "HERE"

            time.sleep(1)

            self.assertTrue(True)
 
    def test_events(self):
        update = None
        def handler(_update, _parent):
            nonlocal update
            update = _update

        service = RawClientEvents(handler, interval=.3)

        time.sleep(1)

        with RealtimeHttpClient("test_sync") as doc:
            u = doc.update

            u.nested.value = "CAsfawojgaw"
            u.name = "TEST"
            u.here = "HERE"

        while update is None:
            time.sleep(.3)

        self.assertIsNotNone(update.get("test_sync",None))

        service.stop()


if __name__ == "__main__":
    unittest.main()



from hashlib import sha256
from pyappi.encoding import encode_url
import os.path

def read1(type, id):
    try:
        with open(f'blocks/{type}/{id}', mode='rb') as file:
            return file.read()
    except Exception as e:
        return

def write1(type, content):
    h256 = sha256()
    h256.update(content)

    id = encode_url(h256.digest())
    
    with open(f'blocks/{type}/{id}', mode='wb') as file:
        file.write(content)
    
    return id

def is1(type,id):
    return {"id": 1 if os.path.isfile(f'blocks/{type}/{id}') else 0 }

def read2(type, id):
    return read1(type, id)

def write2(type, id, content):
    with open(f'blocks/{type}/{id}', mode='wb') as file:
            return file.write(content)

def is2(type,id):
    return is1(type,id)


def type_lookup(name, object = {}):
    type = ""

    if name.find('~') != -1:
        type = name.split("~")[-1]

    return type or object.get("_~","")

def base_key(name):
    return name.split("~")[0]

def split_key(name, object = {}):
    return base_key(name), type_lookup(name, object)


from pathvalidate import sanitize_filepath


def clean_filename(filename):
    filename = sanitize_filepath(filename)
    count = 1
    while count:
        new_filename = filename.replace("..", ".")
        count = filename.count("..")
        filename = new_filename

    return filename


import unittest
from pyappi.client import RealtimeHttpClient, use_test_client
from pyappi.util.login import make_session
from pyappi.document import VolatileDocument, Document


test_documents = [(RealtimeHttpClient,"tally")] #, (VolatileDocument,"vd_tally"), (Document,"d_tally")]


class TestAppiTally(unittest.TestCase):
    def setUp(self):
        use_test_client()
 
        [doc(name).delete() for (doc,name) in test_documents]

    def _test_tally_base(self,_doc_type, _doc_name):
        serial = None
        with _doc_type(_doc_name) as doc:
            doc._perm.stats = "enable"
            doc.obj = {
                "count~tally":17,
                "other~tally":3
            }
            
            doc.obj2_type_tally = {
                "count": 13,
                "other": 1
            }

        with _doc_type(_doc_name) as doc:
            serial = doc["$id"].serial

            self.assertEqual(doc._server_tally.count, 30)
            self.assertEqual(doc._server_tally.other, 4)

            doc.obj = {
                "count~tally":7,
                "other~tally":4,
                "new~tally":2
            }

            doc.obj2_type_tally = {
                "count": 13,
                "other": 1
            }

        with _doc_type(_doc_name) as doc:

            self.assertEqual(doc._server_tally.count, 20)
            self.assertEqual(doc._server_tally.other, 5)
            self.assertEqual(doc._server_tally.new, 2)

            doc.test_type_log = {"_depth":5}

            for i in range(10):
                doc.test_type_log[f'x{i}'] = {"internal~ltally": 1,"message": "message", "index":i}
                doc.flush()

            doc.sync()

            #std::string serial;serial = std::string("!") + std::string(root("$id")["serial"]);
            self.assertEqual(doc._server_ltally.test, 10)
            self.assertEqual(doc._server_ltally.internal, 10)

            doc.subscribers_type_log.x = {
                "content": "I subscribed",
            }

            doc.flush()

            with _doc_type(_doc_name,session=make_session("TESTUSER", "TESTPASSWORD", params={"type":"stats"})) as stats:
                self.assertEqual(stats["subscribers"], 1)

            with _doc_type("@user") as user:
                user.subscriptions_type_glog.x = serial
                user.following_type_glog.x = serial


            with _doc_type(_doc_name,session=make_session("TESTUSER", "TESTPASSWORD", params={"type":"stats"})) as stats:
                self.assertEqual(stats["followers"], 1)
                self.assertEqual(stats["subscribers"], 2)

            with _doc_type(_doc_name,session=make_session("TESTUSER2", "TESTPASSWORD2", params={"type":"comments"})) as comment:
                comment.x = {"message": "Access Denied"}

            self.assertEqual(comment.status_code,403)

            doc._perm.comments = "enable"
            doc.sync()

            with _doc_type(_doc_name,session=make_session("TESTUSER2", "TESTPASSWORD2", params={"type":"comments"})) as comment:
                comment.x = {"message": "Allow Access"}

            self.assertEqual(comment.status_code,202)

            with _doc_type("@user") as user:
                id = user._server_id.serial

            self.assertEqual(user.status_code,200)

            with _doc_type("@user",session=make_session("TESTUSER2", "TESTPASSWORD2")) as user2:
                id2 = user2._server_id.serial

            self.assertEqual(user2.status_code,200)

            with _doc_type(_doc_name,session=make_session("TESTUSER2", "TESTPASSWORD2", params={"type":"comments"})) as comment:
                comment.x = {"message": "message", "~owner": id2}

            self.assertEqual(comment.status_code,202)

            with _doc_type(_doc_name,session=make_session("TESTUSER", "TESTPASSWORD", params={"type":"comments"})) as comment:
                comment.x = {"message": "message", "~owner": id}

            self.assertEqual(comment.status_code,202)

            with _doc_type(_doc_name) as doc:
                for k,v  in doc.comments_type_log.unwrap().items():
                    if k.startswith("_"):
                        continue

                    if v.get("~owner", "") == id:
                        owner_key = k
                    if v.get("~owner", "") == id2:
                        guest_key = k

            with _doc_type(_doc_name,session=make_session("TESTUSER2", "TESTPASSWORD2", params={"type":"comments"})) as comment:
                comment[owner_key] = {"message": "Cant update someone else's message."}

            self.assertEqual(comment.status_code,422)

            with _doc_type(_doc_name,session=make_session("TESTUSER2", "TESTPASSWORD2", params={"type":"comments"})) as comment:
                comment[guest_key] = {"message": "Can update own message"}

            self.assertEqual(comment.status_code,202)

            with _doc_type(_doc_name,session=make_session("TESTUSER", "TESTPASSWORD", params={"type":"comments"})) as comment:
                comment[guest_key] = {"message": "Can update message as owner"}

            self.assertEqual(comment.status_code,202)

            # TODO VALIDATE IO TO HISTORY BUCKETS
            # You should only be able to update log history you own

            """
            // TODO TEST reader interaction permissions
            // readers are people with the _perm: read
            result = client.Upsert(qid, R"J(
            {
                "_perm":{
                    "comments":"readers"
                }
            })J");
            REQUIRE(result == tdb::appi::NoError);"""

    def test_tally(self):
        [self._test_tally_base(doc,name) for (doc,name) in test_documents]


if __name__ == "__main__":
    unittest.main()



from pyappi.document.delta import changes_since
from pyappi.api_base import get_document_type
import os
import json


local_config = {
    "enable": True,
    "user_local_tsx_root": "appidb/user_local_tsx",
}

if local_config["user_local_tsx_root"]:
    os.makedirs(local_config["user_local_tsx_root"],exist_ok=True)

def sync_user_local_transaction(tsx, user_id):
    with get_document_type()(f'local_tsx.{user_id}', user_id,local_config["user_local_tsx_root"]) as doc:
        return changes_since(doc,tsx)
    
def sync_user_local_transaction2(tsx, user_id):
    with get_document_type()(f'local_tsx.{user_id}', user_id,local_config["user_local_tsx_root"]) as doc:
        return changes_since(doc,tsx) if tsx != -1 else {}, doc._cmt

def sync_root(user_id):
    with get_document_type()(f'local_tsx.{user_id}', user_id,local_config["user_local_tsx_root"]) as doc:
        return doc.unwrap()


def update_user_local_transaction(user_id, document_id, new_tsx, is_public):
    name = f'{local_config["user_local_tsx_root"]}/local_tsx.{user_id}.json'
    current = {}
    try:
        with open(name) as document_handle:
            current = json.load(document_handle)
    except Exception as e:
        pass

    tsx = current.get("_cmt",0) + 1

    current[document_id] = {"tsx":new_tsx,"pub":int(is_public),"_vmt":tsx}

    current["_cmt"] = tsx

    try:
        with open(name, "w") as doc:
            doc.write(json.dumps(current , indent=4))
    except Exception as e:
        pass


from pyappi.util.login import session_challenge

session = {}

def get_session():
    global session
    return session

def set_session(user, password, more={}):
    global session
    session = session_challenge(user, password, more)

    return session



from typing import Any
from .random_handler import random_handler
from .type import type_lookup
from .exceptions import *
from .file_handler import file_handler
from pyappi.document.lock import global_appi_mutex
from pyappi.document.transaction import Transaction
from pyappi.document.local import update_user_local_transaction
from pyappi.document.history import update_document_history
from pyappi.util.merge import dict_merge
from pyappi.util.filename import clean_filename
from pyappi.document.enumeration import prepare_new_document
from pyappi.stats.stats import read_stats
from pyappi.document.history import read_history

import json
import os
import time


def fs_enter(self):
    global_appi_mutex.acquire()

    self.__dict__['__lock'] = True
    try:
        mode = self.__dict__['__session'].get("type", None)
        match mode:
            case "stats":
                self.__dict__['__document'] = read_stats(self.name)
            case "history":
                self.__dict__['__document'] = read_history(self.name)
            case "public":
                filename = clean_filename(f'{self.__dict__["__path"] }/{self.name}.json')
                with open(filename) as document_handle:
                    self.__dict__['__document'] = json.load(document_handle)["~public"]
            case _:
                filename = clean_filename(f'{self.__dict__["__path"] }/{self.name}.json')
                with open(filename) as document_handle:
                    self.__dict__['__document'] = json.load(document_handle)
    except Exception as e:
        if self.__dict__['__read_only']:
            self.__dict__['__lock'] = False
            global_appi_mutex.release()
            raise e
        
        self.__dict__['__document'] = {}
        if document_config["enumeration"]:
            prepare_new_document(self,self.name)

    self.__dict__['__tsx'] = self.__dict__['__document'].get("_cmt",0) + 1

    return self


def fs_delete(self):
    record = f'{self.__dict__["__path"]}/{self.__dict__["name"]}.json'
    backup = f'{self.__dict__["__path"]}/{self.__dict__["name"]}.backup.json'

    try:
        os.remove(backup)
    except OSError:
        pass

    try:
        os.remove(record)
    except OSError:
        pass


def fs_sync(self):
    # With the current implementation the read back is a no op.
    fs_flush(self)



def fs_flush(self):
    if not self.__dict__['__mutated'] or self.__dict__["__read_only"]:
        return
    
    record = clean_filename(f'{self.__dict__["__path"]}/{self.__dict__["name"]}.json')
    backup = clean_filename(f'{self.__dict__["__path"]}/{self.__dict__["name"]}.backup.json')

    try:
        os.remove(backup)
    except OSError:
        pass

    try:
        os.rename(record, backup)
    except OSError:
        pass
    
    with open(record, "w") as doc:
        doc.write(json.dumps(self.__dict__['__document'] , indent=4))

    tsx = self.__dict__['__tsx']
    self.__dict__['__tsx'] = self.__dict__['__document'].get("_cmt",0) + 1

    permissions = self.__dict__['__document'].get("_perm",{})
    is_public = permissions.get("public","") == "read"

    for user,level in permissions.items():

        if user == "public" or user[0] == '_' or level == "inherit":
            continue

        update_user_local_transaction(user, self.__dict__["name"], tsx, is_public)

    update_document_history(self.__dict__['__who'],self.__dict__['name'],tsx, is_public, self.__dict__['__delta'])
    self.__dict__['__delta'] = {}
    self.__dict__['__mutated'] = False


def fs_exit(self, type, value, traceback):
    fs_flush(self)

    self.__dict__['__lock'] = False
    global_appi_mutex.release()

document_config = {
    "root": "appidb/documents",
    "enumeration": True,
}

if document_config["root"]:
    os.makedirs(document_config["root"],exist_ok=True)


class Document():
    def __init__(self, name, who="", path=None, read_only=False, auto_nav=True, io={"enter":fs_enter,"delete":fs_delete,"exit":fs_exit, "flush": fs_flush, "sync": fs_sync}, client=False, session={}, who_id=None):
        self.__dict__['__path'] = document_config["root"] if not path else path
        self.__dict__['__auto_nav'] = auto_nav
        self.__dict__['name'] = name
        self.__dict__['__io'] = io
        self.__dict__['__client'] = client
        self.__dict__['__who'] = who
        self.__dict__['__who_id'] = who_id
        self.__dict__['__tsx'] = 0
        self.__dict__['__lock'] = False
        self.__dict__['__mutated'] = False
        self.__dict__['__read_only'] = read_only
        self.__dict__['__document'] = {}
        self.__dict__['__delta'] = {}
        self.__dict__['__now'] = int(time.time() * 1000 * 1000)
        self.__dict__["__session"] = session

        mode = session.get("type", None)
        if mode:
            match mode:
                case "stats" | "history" | "public":
                    self.__dict__['__read_only'] = True
                    pass
                case _:
                    pass


    def __len__(self):
        return len(self.__dict__['__document'])
    
    def __contains__(self, key):
        return key in self.__dict__['__document']

    def unwrap(self):
        return self.__dict__['__document']
    
    def get_id(self):
        return self.__dict__['name']
    
    def get_user(self):
        return self.__dict__['__who']
    
    def get_user_id(self):
        return self.__dict__['__who_id']
    
    def get_document_type(self):
        return Document
    
    def get(self,key,default):
        if self.__dict__['__document'].get(key,None) is None:
            return default
        
        return self[key]

    def __update(self, tsx, key, delta):
        if delta and not (key.startswith("$") and self.__dict__['__client']):
            self.__dict__['__delta'][key] = self.__dict__['__delta'].get(key,{})
            dict_merge(self.__dict__['__delta'][key],delta)

        self.__dict__['__mutated'] = True
        self.__dict__['__document']["_cmt"] = tsx
        self.__dict__['__document']["_lmt"] = int(time.time())

    def __setattr__(self, _name: str, _value: Any) -> None:
        _name = _name.replace("_type_","~").replace("_server_","$")

        if self.__dict__['__read_only']:
            raise ReadOnlyDocument()

        if not self.__dict__['__lock']:
            raise DocumentNotLocked()

        if not isinstance(_value, dict):
            raise "Root values not allowed in Appi"
        
        if len(_value):
            t = self[_name]

            for k,v in _value.items():
                t[k] = v

            return

        self.__dict__['__document'][_name] = _value
        self.__update(self.__dict__['__tsx'],_name, self.__dict__['__document'][_name])

    def __getattr__(self, key):
        key = key.replace("_type_","~").replace("_server_","$")

        if not self.__dict__['__lock']:
            raise DocumentNotLocked()

        if key == '_Transaction__update':
            return self._Document__update
        doc = self.__dict__['__document'].get(key,None)
        if not doc:
            if self.__dict__['__auto_nav'] and not self.__dict__['__read_only']:
                if key.startswith("__") and key.endswith("__"):
                    return None
                self.__dict__['__document'][key] = {}
                doc = self.__dict__['__document'][key]
            else:
                raise PathDoesntExist()
            
        if isinstance(doc, dict):  
            return Transaction(doc, self.__dict__['__tsx'],key, self, type_lookup(key,doc), self)
        else:
            return doc

    def __setitem__(self, __name: str, __value: Any) -> None:
        return self.__setattr__(__name,__value)
    
    def __getitem__(self, _name: str):
        return self.__getattr__(_name)

    def __enter__(self):
        document = self.__dict__['__io']["enter"](self)

        if not self.__dict__["__client"]:
            self.__dict__["__is_owner"] = self.__dict__['__document'].get("_perm",{}).get(self.__dict__['__who'],"") == "owner"
        
            return document

        mode = self.__dict__['__session'].get("type", None)

        match mode:
            case "comment":
                return document.comments_type_log
            case _:
                return document
    
    def delete(self):
        return self.__dict__['__io']["delete"](self)
    
    def flush(self):
        return self.__dict__['__io']["flush"](self)
    
    def sync(self):
        return self.__dict__['__io']["sync"](self)
    
    def __exit__(self, type, value, traceback):
        return self.__dict__['__io']["exit"](self,type,value,traceback)







import httpx
from fastapi.testclient import TestClient
from pyappi.client.config import use_test_config, get_config
from pyappi.client.session import set_session
from pyappi.util.login import session_challenge
from pyappi.util.login import secure_user


client_handle = httpx


def create_user(user, password, body={}):
    client, config = get_connection_details()
    session = session_challenge(user,password)
    response = client.post(f'{config["protocol"]}{config["host"]}/user/create/{secure_user(user)}?{session}', json=body)


def use_test_client():
    global client_handle
    use_test_config()

    import pyappi.endpoints # This line is required to load the FastAPI enpoints
    from pyappi.api_base import app, set_document_type
    from pyappi.document import Document

    set_document_type(Document)

    client_handle = TestClient(app)
    default_session = set_session("TESTUSER","TESTPASSWORD")
    
    client, config = get_connection_details()
    body = {
        "profile":{}
    }

    create_user("TESTUSER2", "TESTPASSWORD2", body)
    create_user("TESTUSER3", "TESTPASSWORD3", body)
    create_user("TESTUSER","TESTPASSWORD", body)

    return client_handle

def set_client(_client):
    global client_handle
    client_handle = _client

    return client_handle

def get_client():
    global client_handle
    return client_handle

def get_connection_details():
    global client_handle
    return client_handle, get_config()


import os
import time
import json
from threading import RLock, Lock
from pyappi.util.filename import clean_filename
from pyappi.util.interval import Interval


stats_config = {
    "enable": True,
    "stats_root": "appidb/stats",
}

if stats_config["stats_root"]:
    os.makedirs(stats_config["stats_root"],exist_ok=True)


sessions = {}

stats_record_mutex = Lock()
session_lock = RLock()

def close_session(session_id):
    target_id, user_id = session_id.split("|")

    with StatsRecord(target_id) as record:
        record["viewers"] -= 1

    with session_lock:
        del sessions[session_id]

def session_cleanup():
    now = int(time.time())
    with session_lock:
        to_close = []
        for session_id, session in sessions.items():
            if session["last_live"] + 60 < now:
                to_close.append(session_id)

        for session_id in to_close:
            close_session(session_id)
            

session_cleanup_interval = None

def use_session_interval():
    global session_cleanup_interval
    session_cleanup_interval = Interval(60,session_cleanup)


def read_stats(name):
    filename = clean_filename(f'{stats_config["stats_root"] }/{name}.json')
    with open(filename) as document_handle:
        return json.load(document_handle)


class StatsRecord():
    def __init__(self, name):
        self.name = name
        self.record = {"views":0,"viewers":0,"likes":0,"followers": 0, "favorites":0,"time":0,"comments":0,"reviews":0, "subscribers": 0}

    def __enter__(self):
        stats_record_mutex.acquire()

        try:
            filename = clean_filename(f'{stats_config["stats_root"] }/{self.name}.json')
            with open(filename) as document_handle:
                self.record = json.load(document_handle)
        except Exception as e:
            pass

        return self.record
    
    def delete(self):
        filename = clean_filename(f'{stats_config["stats_root"] }/{self.name}.json')

        try:
            os.remove(filename)
        except OSError:
            pass
    
    def __exit__(self, type, value, traceback):
        filename = clean_filename(f'{stats_config["stats_root"] }/{self.name}.json')
        backup = clean_filename(f'{stats_config["stats_root"] }/{self.name}.backup.json')

        try:
            os.remove(backup)
        except OSError:
            pass

        try:
            os.rename(filename, backup)
        except OSError:
            pass
        
        try:
            with open(filename, "w") as doc:
                doc.write(json.dumps(self.record , indent=4))
        except Exception as e:
            pass

        stats_record_mutex.release()



def start_viewing_session(target_id,user_id):
    session_id = f"{target_id}|{user_id}"

    if session_id in sessions:
        return update_viewing_sesison(target_id,user_id)
    
    with StatsRecord(target_id) as record:
        record["views"] += 1
        record["viewers"] += 1

    now = int(time.time())
    
    sessions[session_id] = {
        "started": now,
        "last_live": now
    }

    return record

def update_viewing_sesison(target_id,user_id):
    with session_lock:
        session_id = f"{target_id}|{user_id}"

        if not session_id in sessions:
            return start_viewing_session(target_id,user_id)
        
        now = int(time.time())
        
        sessions[session_id] = {
            "last_live": now
        }
    
    with StatsRecord(target_id) as record:
        return record


import json
import time
import os
from pyappi.util.filename import clean_filename


history_config = {
    "enable": True,
    "document_history_root": "appidb/document_history"
}


def read_history(name):
    print("tdo read history", name)

    return {}


if history_config["document_history_root"]:
    os.makedirs(history_config["document_history_root"],exist_ok=True)


def update_document_history(user_id, document_id, new_tsx, is_public,delta):
    if not history_config["enable"]:
        return

    name = clean_filename(f'{history_config["document_history_root"]}/history.{document_id}.json')

    current = {"time":int(time.time()),"user":user_id,"tsx":new_tsx,"pub":is_public,"delta":delta}

    with open(name, "a") as doc:
        doc.write(json.dumps(current , indent=4))


from pyappi.api_base import app,set_document_type
from pyappi.document.document import Document
from pyappi.endpoints import *
import pyappi
from colorama import Fore, Style

import uvicorn

set_document_type(Document)

def main():
    print(f"{Fore.GREEN}PYAPPI {pyappi.__version__}{Style.RESET_ALL}")
    uvicorn.run("pyappi.server_cli:app", port=8099, log_level="info")

if __name__ == "__main__":
    main()


from pyappi.api_base import app
from pyappi.block.endpoints import *
from pyappi.user.session import verify_session
from pyappi.document.local import sync_user_local_transaction, sync_root, sync_user_local_transaction2


@app.get("/sync/status/{tsx}")
@verify_session
async def sync_status(tsx, request: Request):
    result = sync_user_local_transaction(int(tsx),request.state.session["user"])

    return result

@app.get("/sync/updates/{tsx}")
@verify_session
async def sync_updates(tsx, request: Request):
    updates, tsx = sync_user_local_transaction2(int(tsx),request.state.session["user"])

    return {"updates": updates, "tsx": tsx}

@app.get("/sync/root")
@verify_session
async def _sync_root(request: Request):
    result = sync_root(request.state.session["user"])

    return result



from pyappi.client.client_handle import get_connection_details
from pyappi.util.login import encode_session
from pyappi.client.session import get_session
from pyappi.client.config import set_config
import threading
import time


class RawService:
    def __init__(self, handler, name="default", interval=3, config=None, session={}):
        self.name = name
        self.handler = handler
        self.interval = interval

        if config:
            set_config(config)

        self.session = get_session() if not len(session) else encode_session(session)

        self.running = True
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

    def stop(self):
        self.running = False
        self.thread.join()

    def run(self):
        client, config = get_connection_details()

        res = client.get(f'{config["protocol"]}{config["host"]}/service/cursor?{self.session}').json()
        
        tsx = res["cursor"]

        while self.running:
            time.sleep(self.interval)

            res = client.get(f'{config["protocol"]}{config["host"]}/service/read_events/{tsx}?{self.session}').json()

            if res["events"]:
                self.handler(res)

    
class ServiceTail:
    def __init__(self, config=None, session={}, interval=1):
        self.config = config
        self.session = session
        self.interval = interval

    @staticmethod
    def handler(update):
        print(update)

    def __enter__(self):
        self.service = RawService(ServiceTail.handler,session=self.session, interval=self.interval, config=self.config)

        return self
    
    def __exit__(self, type, value, traceback):
        self.service.stop()
    


def encode_url(bin):
    result = ""

    val = 0
    valb = -6

    for c in bin:
        val = (val << 8) + c
        valb += 8

        while valb >= 0:
            result += "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789[]"[(val >> valb) & 0x3F]
            valb -= 6
        

    if valb > -6:
        result += "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789[]"[((val << 8) >> (valb + 8)) & 0x3F]

    while len(result) % 4:
        result += '-'

    return result
	
def decode_url(str):
    result = bytearray()

    T = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789[]"

    val = 0
    valb = -8

    for c in str:
        dx = T.find(c)

        if dx == -1:
            break

        val = (val << 6) + dx
        valb += 6

        if valb >= 0:
            result.append((val >> valb) & 0xFF)
            valb -= 8
    
    return result


import unittest
import time
from pyappi.document import VolatileDocument, Document

test_documents = [(Document,"test")]#, (VolatileDocument,"vdtest")] Volatile needs history to be implemented

class TestAppiLog(unittest.TestCase):
    def setUp(self):
        [doc(name,"test").delete() for (doc,name) in test_documents]

    def _test_log_base(self,_doc_type, _doc_name):
        with _doc_type(_doc_name,"test") as doc:
        
            doc["comments~log"] = {}
            log  = doc["comments~log"]

            log._depth = 8
            log._size = 256
            log._interval = 1000* 60 * 60
            log._mode = -1

            now = int(time.time())

            for i in range(9):
                log[now + i] = { "message": "message"+str(i) }

            self.assertEqual(log.get(now+0,None), None)

            for i in range(1, 9):
                self.assertEqual(log[now+i].message, "message"+str(i))

            long_message = "123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 1234567890"
            log[now+10] = { "message": long_message}

            self.assertEqual(log[now+10].message, long_message)

            history_id = f'_{doc["$id"].serial}.{now - (now % (1000 * 60 * 60))}'

            with _doc_type(history_id,"test") as history:
                hlog = history.comments

                for i in range(1, 9):
                    self.assertEqual(hlog[now+i].message, "message"+str(i))

                self.assertEqual(hlog[now+10].message, long_message)

    def test_log(self):
        [self._test_log_base(doc,name) for (doc,name) in test_documents]



if __name__ == "__main__":
    unittest.main()



def changes_since(document, tsx):

    return changes_since_dict(document.__dict__["__document"], tsx)


def changes_since_dict(document, tsx):
    result = {}

    def _recursive_changes(current):
        cmt,vmt = (current.get("_cmt",-1),current.get("_vmt",-1))

        if cmt <= tsx and vmt <= tsx:
            return None

        delta = {}

        if cmt > tsx:
            for key,value in current.items():
                if not isinstance(value,dict):
                    continue

                _delta = _recursive_changes(value)
                if _delta:
                    delta[key] = _delta

        if vmt > tsx:
            for key,value in current.items():
                if isinstance(value,dict):
                    continue

                delta[key] = value
            
        return delta

    result = _recursive_changes(document)

    return result




from pyappi.api_base import app
from fastapi import Request
from pyappi.service.session import verify_service
from pyappi.stats.stats import update_viewing_sesison
from pyappi.events.events import read_events, read_cursor


@app.get("/service/read_events/{tsx}")
@verify_service
async def read(tsx, request: Request):
    return read_events(int(tsx))


@app.get("/service/cursor")
@verify_service
async def cursor(request: Request):
    return read_cursor()


from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from contextlib import asynccontextmanager

description = """"""
version = "1.1.6"

@asynccontextmanager
async def lifespan(app: FastAPI):
    from pyappi.stats.stats import session_cleanup_interval, use_session_interval

    use_session_interval()

    #print("startup")

    yield
    #print("shutdown")


    if session_cleanup_interval:
        session_cleanup_interval.stop()

app = FastAPI(title="pyappi", description=description, version=version, lifespan=lifespan)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

document_type = None
def set_document_type(dt):
    global document_type

    document_type = dt

def get_document_type():
    return document_type


import json
import os


enumeration_config = {
    "enable": True,
    "enumeration_root": "appidb/document_enumeration"
}


if enumeration_config["enumeration_root"]:
    os.makedirs(enumeration_config["enumeration_root"],exist_ok=True)

filename = f'{enumeration_config["enumeration_root"] }/serial_to_unique.json'
backup = f'{enumeration_config["enumeration_root"] }/serial_to_unique.backup.json'

def _get_lookup():
    lookup = {}
    try:
        with open(filename) as fh:
            lookup = json.load(fh)
    except Exception as e:
        pass

    return lookup

def lookup_document_id(serial):
    lookup = _get_lookup()

    return lookup.get(str(serial),"")

def is_new_document(document):
    return not len(document["auth"])

def enumerate_document_id(document_id):

    lookup = _get_lookup()

    try:
        os.remove(backup)
    except OSError:
        pass

    try:
        os.rename(filename, backup)
    except OSError:
        pass

    enumerator = len(lookup)
    lookup[enumerator] = document_id

    with open(filename, "w") as doc:
        doc.write(json.dumps(lookup, indent=4))

    return enumerator


def prepare_new_document(document, document_id):
    if not enumeration_config["enable"]:
        return
    
    document["$id"].serial = enumerate_document_id(document_id)
    document["$id"].unique = document_id


def log_handler(serial_id, path, key, time_group, value, parent):
    log_id = f'_{serial_id}.{time_group}'
    doc_type = parent.get_document_type()
    user = parent.get_user()
    user_id = parent.get_user_id()

    with doc_type(log_id,user, who_id=user_id) as doc:
        doc._perm[parent.get_id()] = "inherit"
        doc[path][key] = value


import unittest
from pyappi.client import use_test_client, RealtimeHttpClient
from pyappi.service.service import RawService, ServiceTail
from pyappi.service.session import lookup_service_key
import time


class TestEvents(unittest.TestCase):
    def setUp(self):
        use_test_client()

        self.key = lookup_service_key()
        RealtimeHttpClient("test_events").delete()

    def test_tail(self):
        with ServiceTail(session={"user":"default","challenge":self.key}, interval=.3):

            time.sleep(1)

            with RealtimeHttpClient("test_events") as doc:
                u = doc.update

                u.nested.value = "CAsfawojgaw"
                u.name = "TEST"
                u.here = "HERE"

            time.sleep(1)

            self.assertTrue(True)
 
    def test_events(self):
        update = None
        def handler(_update):
            nonlocal update
            update = _update["events"]

        service = RawService(handler,session={"user":"default","challenge":self.key}, interval=.3)

        time.sleep(1)

        with RealtimeHttpClient("test_events") as doc:
            u = doc.update

            u.nested.value = "CAsfawojgaw"
            u.name = "TEST"
            u.here = "HERE"

        while update is None:
            time.sleep(.3)

        key = list(update)[-1]

        self.assertEqual(update[key]["id"],"test_events")
        self.assertEqual(update[key]["updates"]["update.nested.value"],"set")
        self.assertEqual(update[key]["updates"]["update.name"],"set")
        self.assertEqual(update[key]["updates"]["update.here"],"set")

        service.stop()


if __name__ == "__main__":
    unittest.main()



from hashlib import sha256,sha512
from pyappi.encoding.url import encode_url, decode_url
import json


def secure_user(user):
    h256 = sha256()
    h256.update((user + 'APPI-DOMAIN-06036F34-78A8-45A7-B96D-646BE001039C').encode())

    return encode_url(h256.digest())

def secure_challenge(challenge, user):
    h512 = sha512()
    h512.update((challenge + user + 'APPI-DOMAIN-C3A09828-1AEB-4486-9BDE-010A035FE92C').encode())

    return encode_url(h512.digest())

def session_challenge(_user="", _challenge="", params = {}):
    user = secure_user(_user) if _user else ""
    challenge = secure_challenge(_challenge, user) if _challenge else ""
    session = {}
    session.update({"user": user, "challenge": challenge})
    session.update(params)
    bin = json.dumps(session).encode()

    return encode_url(bin)

def make_session(_user="", _challenge="", params = {}):
    user = secure_user(_user) if _user else ""
    challenge = secure_challenge(_challenge, user) if _challenge else ""
    session = {}
    session.update({"user": user, "challenge": challenge})
    session.update(params)
    
    return session

def encode_session(session):
    bin = json.dumps(session).encode()

    return encode_url(bin)

def decode_session(session):
    return json.loads(decode_url(session).decode())


class DocumentNotLocked(Exception):
    "Raised when access to a document is attempted outside of its lifetime."
    pass

class PathDoesntExist(Exception):
    "Raised when access to a document is attempted outside of its lifetime."
    pass

class ReadOnlyDocument(Exception):
    "Raised when access to a document is attempted outside of its lifetime."
    pass

class UserWriteProtectedRegion(Exception):
    "Raised when access to a document is attempted outside of its lifetime."
    pass

class ListsAreNotAllowed(Exception):
    "Lists just make things harder"
    pass

class DictionaryReassignmentProhibited(Exception):
    "In some cases we don't allow this"
    pass

class NoPrepopulatedObjects(Exception):
    "Raised when assigned objects have values or subkeys, as this bypasses the preprocessor."
    pass

class ValueReassignmentToDictProhibited(Exception):
    "Once a value always a value"
    pass

class BulkAssignmentProhibited(Exception):
    "When the flag is enabled, bulk dict assignment can be prohibitied"
    pass


from pyappi.document.type import *
from pyappi.document.random_handler import *
from pyappi.document.exceptions import *
from pyappi.logs import log_handler
from pyappi.stats import stats_handler
from typing import Any

import time
import os
import json


transaction_config = {
    "allow_bulk_assign": True
}
    

class Transaction:
    def __init__(self, document, tsx, key, parent=None, type="", root=None, owner=None):
        self.__dict__['__document'] = document
        self.__dict__['__parent'] = parent
        self.__dict__['__root'] = root
        self.__dict__['__tsx'] = tsx
        self.__dict__['__type'] = type
        self.__dict__['__key'] = key
        self.__dict__['__owner'] = owner

    def __len__(self):
        return len(self.__dict__['__document'])
    
    def __contains__(self, key):
        return key in self.__dict__['__document']
    
    def unwrap(self):
        return self.__dict__['__document']
    
    def get(self,key,default):
        if self.__dict__['__document'].get(key,None) is None:
            return default
        
        return self[key]

    def __update(self, tsx, key, _delta, path, action):
        self.__dict__['__document']["_cmt"] = self.__dict__['__tsx']

        # Keys that start with $ should not be sent to the server. They are computed for the benefit of the local object only
        delta = None if not _delta or (key.startswith("$") and self.__dict__["__root"].__dict__['__client']) else {key:_delta}

        self.__dict__['__parent'].__update(self.__dict__['__tsx'], self.__dict__['__key'],delta, (key + "." + path) if path else key, action)

    def __setattr__(self, _name: str, _value: Any) -> None:
        _name = str(_name)
        _name = _name.replace("_type_","~").replace("_server_","$")

        if self.__dict__['__root'].__dict__['__read_only']:
            raise ReadOnlyDocument()
        
        if self.__dict__['__document'].get("~owner", None) != None and self.__dict__['__document']['~owner'] != self.__dict__['__root'].__dict__["__who_id"] and not self.__dict__['__root'].__dict__["__is_owner"]:
            raise UserWriteProtectedRegion()

        if isinstance(_value, list):
            raise ListsAreNotAllowed()
        
        if isinstance(self.__dict__['__document'].get(_name,None), dict):
            raise DictionaryReassignmentProhibited()
        
        def ltally(key, count):
            lt = self.__dict__['__root']["$ltally"]

            lt[key] = lt.get(key,0) + count

        def tally(key, count):
            lt = self.__dict__['__root']["$tally"]

            lt[key] = lt.get(key,0) + count
            
        is_client = self.__dict__["__root"].__dict__['__client']
        parent_type = self.__dict__['__type']

        if not is_client:
            match parent_type:
                case "log" | "glog":
                    now = self.__dict__["__root"].__dict__['__now']
                    self.__dict__["__root"].__dict__['__now'] += 1

                    if not self.__dict__['__document'].get("_frame",None):
                        self.__dict__['__document']["_frame"] = now
                    frame = self.__dict__['__document']["_frame"]

                    if not self.__dict__['__document'].get("_depth",None):
                        self.__dict__['__document']["_depth"] = 256
                    depth = self.__dict__['__document']["_depth"]

                    if not self.__dict__['__document'].get("_size",None):
                        self.__dict__['__document']["_size"] = 32*1024
                    size = self.__dict__['__document']["_size"]

                    if not self.__dict__['__document'].get("_interval",None):
                        self.__dict__['__document']["_interval"] = 60*60*24
                    interval = self.__dict__['__document']["_interval"]

                    if not self.__dict__['__document'].get("_mode",None):
                        self.__dict__['__document']["_mode"] = 1 # Enable server time mode
                    time_type = self.__dict__['__document']["_mode"]

                    if _name.startswith("_"):
                        exists = self.get(_name,None) is not None
                        self.__dict__['__document'][_name] = _value
                    else:
                        if isinstance(_value,dict):
                            _value["_size"] = len(json.dumps(_value))

                        if time_type == -1: # Client defined key
                            ikey = int(_name)
                            time_group = str(ikey - (ikey % interval))
                        else:
                            nowr = now - frame
                            _name = str(nowr)
                            time_group = str(nowr - (nowr % interval))
                            
                        base_key = self.__dict__['__key'].split("~")[0]
                        if parent_type == "glog":
                            stats_handler(self.__dict__['__root']["$id"]["unique"], base_key, _value)
                        else:
                            ltally(base_key, 1)
                            stats_handler(self.__dict__['__root']["$id"]["unique"], base_key)
                
                        log_handler(self.__dict__['__root']["$id"]["serial"], base_key, _name, time_group, _value, self.__dict__['__root'])
                    
                        self.__dict__['__type'] = ""
                        exists = self.get(_name,None) is not None
                        self[_name] = _value
                        self.__dict__['__type'] = parent_type

                        items = [k for (k,v) in self.__dict__['__document'].items() if not k.startswith("_")]
                        items.sort()

                        if len(items) > depth:
                            del self.__dict__['__document'][items.pop(0)]

                        current_size = 0
                        delete_index = 0
                        for item in items:
                            if not isinstance(self.__dict__['__document'][item],dict):
                                continue

                            current_size += self.__dict__['__document'][item]["_size"]
                            while current_size > size:
                                current_size -= self.__dict__['__document'][items[delete_index]]["_size"]
                                del self.__dict__['__document'][items[delete_index]]
                                delete_index += 1

                    self.__dict__['__document']["_vmt"] = self.__dict__['__tsx']
                    #self.__update(self.__dict__['__tsx'],_name, self.__dict__['__document'][_name])
                    self.__dict__['__parent'].__update(self.__dict__['__tsx'], self.__dict__['__key'],{_name:self.__dict__['__document'][_name]}, _name, "log" if not exists else "update_log")

                    return
                case _:
                    pass
        
        if isinstance(_value,dict) and len(_value):
            if not transaction_config["allow_bulk_assign"]:
                raise BulkAssignmentProhibited()
            
            if not self.__dict__['__document'].get(_name,None):
                self.__dict__['__document'][_name] = {}
            elif not isinstance(self.__dict__['__document'][_name],dict):
                raise ValueReassignmentToDictProhibited()

            sub = self[_name]
            for k,v in _value.items():
                sub[k] = v

            return

        # Value Assignment:

        base_key, value_type = split_key(_name)
        exists = self.get(_name,None) is not None

        match parent_type:
            case "tally" | "tly" | "ltally" | "lly" | "gly" | "gtally":
                value_type = parent_type
        
        match value_type:
            case "rng" | "random":
                self.__dict__['__document'][_name] = random_handler(_value)

            case "app" | "append":
                base = self.__dict__['__document'].get(_name,"")
                self.__dict__['__document'][_name] = base + _value
                
            case "pre" | "prepend":
                base = self.__dict__['__document'].get(_name,"")
                self.__dict__['__document'][_name] = _value + base
                
            case "flt" | "float":
                base = self.__dict__['__document'].get(_name,0)
                self.__dict__['__document'][_name] = base + _value
                
            case "cnt" | "counter":
                base = self.__dict__['__document'].get(_name,0)
                self.__dict__['__document'][_name] = base + int(_value)

            case "file" | "blocks" | "block" | "folder" | "tfile" | "tblocks" | "tblock" | "tfolder":
                file_handler()
            
            case "ltally" | "lly" | "local_tally":
                """
                    Collects all K:V tally over time by tracking and summing edits only.
                    Doesn't support deletion as it is only computed on mutation and accumulated.
                    Support logs.
                    Local tallys named "like","comments"... etc will automatically be propagated to the stats
                    {
                        something~log:{
                            25151:{files~lly:3}
                        }
                        else~ltally:{
                            files:4
                        },
                        $ltally:
                        {
                            files:7
                        }
                    }
                """

                ltally(base_key, _value - self.__dict__['__document'].get(_name,0))
                self.__dict__['__document'][_name] = _value


            case "tally" | "tly":
                """
                    Collects all current K:V spread throughout the file and sums them in a root key $tally.
                    Supports deletion as is recomputed completely every time.
                    Doesn't support logs as can't recompute the removed log items.
                    {
                        something:{
                            files~tly:3
                        }
                        else~tally:{
                            files:4
                        },
                        $tally:
                        {
                            files:7
                        }
                    }
                """

                # TODO, honor this concept when value is deleted.

                tally(base_key, _value - self.__dict__['__document'].get(_name,0))
                self.__dict__['__document'][_name] = _value
                
            case _:
                self.__dict__['__document'][_name] = _value
            
        self.__dict__['__document']["_vmt"] = self.__dict__['__tsx']
        self.__dict__['__parent'].__update(self.__dict__['__tsx'], self.__dict__['__key'],{_name:self.__dict__['__document'][_name]}, _name, "set" if not exists else "update")
        #self.__update(self.__dict__['__tsx'],_name, self.__dict__['__document'][_name])

    def __getattr__(self, _name: str):
        _name = str(_name)
        _name = _name.replace("_type_","~").replace("_server_","$")
        _value = self.__dict__['__document'].get(_name,None)

        if _value is None and self.__dict__['__root'].__dict__['__auto_nav'] and not self.__dict__['__root'].__dict__['__read_only']:
            if _name.startswith("__") and _name.endswith("__"):
                return None
            self.__dict__['__document'][_name] = {}
            _value = self.__dict__['__document'][_name]

        if isinstance(_value, dict):
            return Transaction(_value, self.__dict__['__tsx'], _name, self, type_lookup(_name,_value), self.__dict__["__root"], self.__dict__["__owner"] if self.__dict__["__owner"] else self.__dict__["__document"].get("~owner", None))

        return _value
    
    def __setitem__(self, _name: str, _value: Any) -> None:
        return self.__setattr__(_name,_value)
    
    def __getitem__(self, _name: str):
        return self.__getattr__(_name)


from pyappi.encoding.url import decode_url
from pyappi.api_base import get_document_type
import json
from fastapi import Response
from functools import wraps


def verify_session(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        request = kwargs["request"]
        _session = get_user_session(request)
        if not _session:
            return Response(status_code=403)
    
        session, user = _session
        request.state.session = session
        request.state.user = user

        return await func(*args, **kwargs)

    return wrapper

def parse_session(blob):
    return json.loads(decode_url(blob).decode())

def validate_user_session(request):
    raw_session = request.query_params._list[0][0]
    session = parse_session(raw_session)
    
    with get_document_type()(f'user.{session["user"]}', session["user"], read_only=True) as doc:
        return session["challenge"] == doc.auth.challenge
    
    return False

def get_user_session(request):
    raw_session = request.query_params._list[0][0]
    session = parse_session(raw_session)
    
    try:
        with get_document_type()(f'user.{session["user"]}', session["user"], read_only=True) as doc:
            session["id"] = doc._server_id.serial
            return (session,doc) if session["challenge"] == doc.auth.challenge else None
    except Exception as _e:
        pass
    
    return None


import collections

def dict_merge(dct, merge_dct):
    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], dict)): 
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


import pyappi
from colorama import Fore, Style
import httpx
import click
import json
import os

from pyappi.util.login import session_challenge


def resolve_session(user, password):
    if user and password:
        return session_challenge(user, password)
    
    if os.path.exists("appi_session.b64"):
        with open("appi_session.b64","r") as file:
            return file.read()
        
    return ""

def resolve_config(host, proto):
    if host and proto:
        return (host, proto)

    try:
        with open("appi_config.json","r") as file:
            cfg = json.load(file)

            return (cfg["host"], cfg["proto"])
    except Exception as e:
        pass


@click.command()
@click.argument('cmd')
@click.option("--id", default="", help="Id of resource to read")
@click.option("--proto", default="http", help="Protocol string, usually http or https")
@click.option("--host", default="127.0.0.1:8099", help="Host name or ip address with optional port after a :")
@click.option("--user", default="", help="User name of account you are logging into.")
@click.option("--password", default="", help="Password of your account.")
@click.option("--data", default="", help="Json string of data for the specified command")
def read(cmd, id, proto, host, user, password, data):
    (host, proto) = resolve_config(host,config)

    result = httpx.get(f"{proto}://{host}/document/{id}?{resolve_session(user,password)}")
    print(cmd, id, result,result.status_code)


@click.command()
@click.argument('cmd')
@click.option("--id", default="", help="Id of resource to read")
@click.option("--proto", default="https", help="Protocol string, usually http or https")
@click.option("--host", default="appi.host/api", help="Host name or ip address with optional port after a :")
@click.option("--user", default="", help="User name of account you are logging into.")
@click.option("--password", default="", help="Password of your account.")
@click.option("--data", default="", help="Json string of data for the specified command")
def write(cmd, id, proto, host, user, password, data):
    (host, proto) = resolve_config(host,config)

    write = json.parse(data)
    result = httpx.put(f"{proto}://{host}/document/{id}?{resolve_session(user,password)}", json=write)
    print(cmd, id, result,result.status_code)


@click.command()
@click.argument('cmd')
@click.option("--id", default="", help="Id of resource to read")
@click.option("--proto", default="https", help="Protocol string, usually http or https")
@click.option("--host", default="appi.host/api", help="Host name or ip address with optional port after a :")
@click.option("--user", default="", help="User name of account you are logging into.")
@click.option("--password", default="", help="Password of your account.")
@click.option("--data", default="", help="Json string of data for the specified command")
def utail(cmd, id, proto, host, user, password, data):
    from pyappi.client.sync import UserTail
    import time

    session = resolve_session(user,password)
    (host, proto) = resolve_config(host,config)
    
    tail = UserTail(session=session,config={"host":host,"protocol":proto})
    time.sleep(9999999)


@click.command()
@click.argument('cmd')
@click.option("--id", default="", help="Id of resource to read")
@click.option("--proto", default="https", help="Protocol string, usually http or https")
@click.option("--host", default="appi.host/api", help="Host name or ip address with optional port after a :")
@click.option("--user", default="", help="User name of account you are logging into.")
@click.option("--password", default="", help="Password of your account.")
@click.option("--data", default="", help="Json string of data for the specified command")
def stail(cmd, id, proto, host, user, password, data):
    print("TODO")


@click.command()
@click.option("--user", default="", help="User name of account you are logging into.")
@click.option("--password", default="", help="Password of your account.")
def login(user, password):
    session = session_challenge(user, password)

    with open("appi_session.b64","w") as file:
        file.write(session)

@click.command()
@click.option("--proto", default="https", help="Protocol string, usually http or https")
@click.option("--host", default="appi.host/api", help="Host name or ip address with optional port after a :")
def config(host, proto):
    with open("appi_config.json","w") as file:
        file.write(json.dumps({
            "host": host,
            "proto": proto
        }))

@click.command()
def logout():
    try:
        os.remove("appi_session.b64")
    except Exception as _e:
        pass

@click.command()
@click.argument('cmd')
@click.option("--id", default="", help="Id of resource to read")
@click.option("--proto", default="http", help="Protocol string, usually http or https")
@click.option("--host", default="", help="Host name or ip address with optional port after a :")
@click.option("--user", default="", help="User name of account you are logging into.")
@click.option("--password", default="", help="Password of your account.")
@click.option("--data", default="", help="Json string of data for the specified command")
def main(cmd, id, proto, host, user, password):
    match cmd:
        case "read":
            return read()
        case "utail":
            return utail()
        case "stail":
            return stail()
        case "write":
            return write()
        case "login":
            return login()
        case "config":
            return config()
        case "logout":
            return logout()
        case "about":
            print(f"{Fore.GREEN}PYAPPI-CLIENT {pyappi.__version__}{Style.RESET_ALL}")

if __name__ == "__main__":

    
    main()


from pyappi.api_base import app
from pyappi.block.endpoints import *
from pyappi.document.endpoints import *
from pyappi.user.endpoints import *
from pyappi.sync.endpoints import *
from pyappi.stats.endpoints import *
from pyappi.service.endpoints import *


@app.get("/pyappi")
async def pyappi_get_document():
    return {"message": "Hello pyappi World"}

@app.get("/stats")
async def get_stats():
    return {"message": "wouldn't you like to know"}




import unittest
from pyappi.document import VolatileDocument, Document
from pyappi.document.exceptions import DocumentNotLocked


test_documents = [(VolatileDocument,"vdtest"), (Document,"dtest")]


class TestAppiCore(unittest.TestCase):
    def setUp(self):
        [doc(name).delete() for (doc,name) in test_documents]

    def _test_transaction_base(self,_doc_type, _doc_name):
        with _doc_type(_doc_name) as doc:

            doc.types = {}
            doc.types.int = 1
            doc.types.string = "string"
            doc.types.float = 63.437

            self.assertEqual(1,doc.types.int)
            self.assertEqual("string",doc.types.string)
            self.assertEqual(63.437,doc.types.float)


            doc.types.subobj = {}
            doc.types.subobj["test~pre"] = "after"
            doc.types.subobj["test~pre"] = "before-"

            self.assertEqual("before-after",doc.types.subobj["test~pre"])

        self.assertRaises(DocumentNotLocked, lambda: doc.types)
        self.assertRaises(DocumentNotLocked, lambda: setattr(doc.types,"int",2))
            

    def test_transaction(self):
        [self._test_transaction_base(doc,name) for (doc,name) in test_documents]



if __name__ == "__main__":
    unittest.main()



from pyappi.stats.stats import StatsRecord
from pyappi.document.enumeration import lookup_document_id

def stats_handler(target_id, metric, serial_id=None):
    if serial_id:
        # Require global tallies on the user object
        if not target_id.startswith("user."):
            return

        target_id = lookup_document_id(serial_id)

    # Translate between first and third person.
    match metric:
        case "following":
            metric = "followers"
        case "subscriptions":
            metric = "subscribers"
        case "views":
            metric = "view"

    match metric:
        case "likes" | "comments" | "reviews" | "followers" | "subscribers" | "favorites" | "view" | "time":
            with StatsRecord(target_id) as record:
                record[metric] += 1
        case _:
            pass


from pyappi.client.client_handle import get_connection_details
from pyappi.util.login import encode_session
from pyappi.client.session import get_session
from pyappi.client.config import set_config
import threading
import time


class RawClientEvents:
    def __init__(self, handler,interval=3, config=None, session={}):
        self.handler = handler
        self.interval = interval

        if config:
            set_config(config)

        self.session = get_session() if not len(session) else encode_session(session)

        self.running = True
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

    def stop(self):
        self.running = False
        self.thread.join()

    def run(self):
        client, config = get_connection_details()

        _res = client.get(f'{config["protocol"]}{config["host"]}/sync/updates/-1?{self.session}')
        res = _res.json()
        
        tsx = res["tsx"]

        while self.running:
            time.sleep(self.interval)

            try:
                res = client.get(f'{config["protocol"]}{config["host"]}/sync/updates/{tsx}?{self.session}').json()

                tsx = res["tsx"]

                if res["updates"]:
                    self.handler(res["updates"], self)
            except Exception as e:
                time.sleep(5)


class UserTail:
    def __init__(self, config=None, session={}, interval=1):
        self.config = config
        self.session = session
        self.interval = interval

    @staticmethod
    def handler(update, parent):
        print(update)

    def __enter__(self):
        self.client = RawClientEvents(UserTail.handler,self.interval,self.config,self.session)

        return self
    
    def __exit__(self, type, value, traceback):
        self.client.stop()


class UserChanges:
    def __init__(self, config=None, session={}, interval=1):
        self.config = config
        self.session = session
        self.interval = interval

    @staticmethod
    def handler(update, parent):
        for document_name in update:
            client, config = get_connection_details()

            tsx = update[document_name]["tsx"] - 1

            _res = client.get(f'{config["protocol"]}{config["host"]}/document/delta/{document_name}/{tsx}?{parent.session}')
            res = _res.json()

            print(document_name, res)

    def __enter__(self):
        self.client = RawClientEvents(UserChanges.handler,self.interval,self.config,self.session)

        return self
    
    def __exit__(self, type, value, traceback):
        self.client.stop()


from functools import wraps
from fastapi import Response
from pyappi.user.session import parse_session
import os
import json
import base64

service_config = {
    "root": "appidb/service",
}

if not os.path.exists(service_config["root"]):
    os.makedirs(service_config["root"],exist_ok=True)

    try:
        with open(service_config["root"] + "/services.json", "w") as doc:
            key = base64.b64encode(os.urandom(32)).decode()
            doc.write(json.dumps({
                "services":{
                    "default":key
                }
            } , indent=4))
    except Exception as e:
        pass

def lookup_service_key(name="default"):
    try:
        with open(service_config["root"] + "/services.json", "r") as doc:
            services = json.load(doc)["services"]

            return services.get(name,None)
    except Exception as _e:
        return None

def get_service_session(request):
    raw_session = request.query_params._list[0][0]
    session = parse_session(raw_session)

    key = lookup_service_key(session["user"])
    
    return session, session["user"] if key == session["challenge"] else None


def verify_service(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        request = kwargs["request"]
        _session = get_service_session(request)
        if not _session:
            return Response(status_code=403)
    
        session, service = _session
        request.state.session = session
        request.state.user = service

        return await func(*args, **kwargs)

    return wrapper


from pyappi.api_base import app, get_document_type
from fastapi import Request, Response
from pyappi.user.session import parse_session, validate_user_session, verify_session
from pyappi.document.enumeration import is_new_document, prepare_new_document

@app.get("/user/token/{id}")
async def generate_user_token(id, request: Request):
    return {}

@app.post("/user/create/{id}")
async def create_user(id, request: Request):
    raw_session = request.query_params._list[0][0]
    session = parse_session(raw_session)
    body = await request.json()
    
    document_id = f'user.{id}'

    with get_document_type()(document_id, session["user"]) as doc:
        if not is_new_document(doc):
            return Response(status_code=409)
        
        # prepare_new_document(doc,document_id)
        
        doc._perm[id] = "owner"
        doc.auth.challenge = session["challenge"]
        doc.auth.enc_pri_key = body.get("enc_pri_key","")
        doc["~public"].identity.pub_key = body.get("pub_key","")
        doc["~public"].profile = body.get("profile",{})

    with get_document_type()(f'subs.{id}', session["user"]) as doc:
        if len(doc._perm):
            return Response(status_code=409)
        
        doc._perm[id] = "owner"

    return Response(status_code=201)

@app.post("/user/validate/{id}")
@verify_session
async def validate_user(id, request: Request):
    return Response(status_code=202)


from fastapi import WebSocket
from pyappi.api_base import app


# https://fastapi.tiangolo.com/advanced/websockets/
@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    while True:
        data = await websocket.receive_text()
        await websocket.send_text(f"Message text was: {data}")


from pyappi.api_base import app
from fastapi import Request
from pyappi.user.session import verify_session
from pyappi.stats.stats import update_viewing_sesison


@app.post("/stats/view/{id}/{session}/{secret}")
@verify_session
async def view(id, session, secret, request: Request):
    stats = update_viewing_sesison(id, request.state.session["user"])

    return {"session":0, "secret": 0, "stats": stats}


from pyappi.api_base import get_document_type as dt
from fastapi import Response


def document_with_write_access(document_id, session, callback):
    user = session["user"]
    try:
        with dt()(document_id, user, who_id=session["id"]) as doc:
            permissions = doc._perm.unwrap()

            if not len(permissions):
                doc._perm[user] = "owner"
                return callback(doc)

            level = permissions.get("public",None)

            if user in permissions:
                level = permissions[user]
                
            match session.get("type",""):
                case "comments":
                    return callback(doc.comments_type_log) if permissions.get("comments","") == "enable" else Response(status_code=403)
                case _:
                    pass	

            match level:
                case "write" | "owner":
                    return callback(doc)
                case _: 
                    return Response(status_code=409)
    except Exception as e:
        return Response(status_code=422)
    
def document_with_owner_access(document_id, session, callback):
    user = session["user"]
    try:
        with dt()(document_id, user) as doc:
            permissions = doc._perm.unwrap()

            if not len(permissions):
                return callback(doc)

            level = permissions.get("public",None)

            if user in permissions:
                level = permissions[user]

            match level:
                case "owner":
                    return callback(doc)
                case _: 
                    return Response(status_code=409)
    except Exception as e:
        return Response(status_code=422)


def document_with_read_access(document_id, session):
    user = session["user"]
    try:
        
        with dt()(document_id, user) as base, dt()(document_id, user, session=session) as doc:
            permissions = base._perm.unwrap()

            match session.get("type",""):
                case "stats":
                    return doc.unwrap() if permissions.get("stats","") == "enable" else {}
                case "comments":
                    return {"comments~log":doc.comments_type_log.unwrap()} if permissions.get("comments","") == "enable" else {}
                case _:
                    pass

            level = permissions.get("public",None)

            if user in permissions:
                level = permissions[user]

            match level:
                case "read" | "write" | "owner" | "unlisted":
                    return doc.unwrap()
                case _: 
                    return {}
    except Exception as e:
        return None

    """			std::string Get(std::string_view _real_id, const Session& session)
			{
				auto [real_id, query] = d8u::util::split_pair(_real_id, "@");

				auto *record = GetRecord(real_id);
				std::string_view p_cap;

				if (record)
				{
					auto permissions = record->index("_perm");
					if (session.requested_permission.starts_with("child_"))
					{
						// Proxy to child resource
						auto child_resource_name = std::string_view(session.requested_permission.data() + 6, session.requested_permission.size() - 6);
						d8u::util::replace_char(child_resource_name, ',', '.');

						//auto children = record->index("_children");
						//auto child_id = children[child_resource_name];
						auto* child = GetRecord(child_resource_name);

						if (!child)
							return "";

						if (child->index("_perm")[real_id] != "inherit")
							p_cap = std::string_view(child->index("_perm")[real_id]);

						//TODO Fix query by replacing real_id@query with child_id@query
						record = child;
					}
					else if (real_id[0] == '_')
					{
						auto nid = std::stoull(real_id.data()+1);
						auto* parent = GetRecord(std::string("!") + std::to_string(nid));

						if (!parent)
							return "";

						if (!session.HaveReadAccess(parent->index("_perm")))
							return "";

						return record->data;
					}
					else
					{
						switch (switch_t(session.requested_permission)) {
						case switch_t("member"):
							if (session.app != "APPIUSERS")
								return "";

							return "{}"; // TODO SHARE RELEVANT USER DATA WITH MEMBER APPS
						case switch_t("public"):
						{
							auto _public = record->index("~public");

							// Return the public region if it exists
							if (_public.Valid())
								return std::string(_public.Json());

							// Check if this is a public resource and return that
							if (session.HaveReadAccess(permissions))
								return record->data;

							return "";
						}
						case switch_t("share"):
						{
							auto share = record->index("~share");
							auto user_mask = share("access")(session.user);
							auto public_mask = share("access")("public");

							if (user_mask.Valid())
								return Mask(share("data"), user_mask);
							else if (public_mask.Valid())
								return Mask(share("data"), public_mask);

							return "";
						}
						default:
							break;
						}
					}

					if (!session.HaveReadAccess(permissions, p_cap))
						return "";

					if (query.size())
						return provider.Get(_real_id);

					return record->data;
				}

				return "";
			}"""


config = {}

def use_test_config():
    global config
    config = {
        "protocol":"",
        "host":""
    }

    return config

def use_local_config():
    global config
    config = {
        "protocol":"http://",
        "host":"127.0.0.1"
    }

    return config

def set_config(_config):
    global config
    config = _config

    return config

def get_config():
    return config


from pyappi.client.client_handle import get_connection_details
from pyappi.client.session import get_session
from pyappi.client.lock import client_appi_mutex
from pyappi.encoding.url import decode_url
from pyappi.document import Document
from pyappi.util.login import encode_session, decode_session
from pyappi.client.config import set_config

import json


def _resolve_session(self):
    return get_session() if not len(self.__dict__["__session"]) else encode_session(self.__dict__["__session"])

def http_enter(self):
    client_appi_mutex.acquire()

    client, config = get_connection_details()
    session = _resolve_session(self)

    match self.name:
        # TODO Move to helper function
        case "@user":
            self.__dict__["name"] = f'user.{decode_session(session)["user"]}'
        case _: pass

    self.__dict__['__lock'] = True
    try:
        response = client.get(f'{config["protocol"]}{config["host"]}/document/{self.name}?{session}')

        self.__dict__["status_code"] = response.status_code
        if response.status_code != 200:
            raise Exception()
        self.__dict__['__document'] = response.json()
    except Exception as e:
        if self.__dict__['__read_only']:
            self.__dict__['__lock'] = False
            client_appi_mutex.release()
            raise e
        
        self.__dict__['__document'] = {}

    self.__dict__['__tsx'] = self.__dict__['__document'].get("_cmt",0) + 1

    return self


def http_sync(self):
    if self.__dict__['__lock'] != True:
        raise "Document not locked"
    
    if len(self.__dict__['__delta']):
        http_flush(self)

    client, config = get_connection_details()
    session = _resolve_session(self)

    try:
        response = client.get(f'{config["protocol"]}{config["host"]}/document/{self.name}?{session}')
        self.__dict__["status_code"] = response.status_code
        self.__dict__['__document'] = response.json()
    except Exception as e:
        pass


def http_delete(self):
    client, config = get_connection_details()
    session = _resolve_session(self)

    response = client.delete(f'{config["protocol"]}{config["host"]}/document/{self.name}?{session}')
    self.__dict__["status_code"] = response.status_code


def http_flush(self):
    if not self.__dict__['__mutated'] or self.__dict__["__read_only"]:
        return
    
    client, config = get_connection_details()
    session = _resolve_session(self)

    self.__dict__['__delta']["_bmt"] = -1
    response = client.put(f'{config["protocol"]}{config["host"]}/document/{self.name}?{session}', json=self.__dict__['__delta'])
    self.__dict__['__delta'] = {}
    self.__dict__['__mutated'] = False
    self.__dict__["status_code"] = response.status_code


def http_exit(self, type, value, traceback):
    http_flush(self)

    self.__dict__['__lock'] = False
    client_appi_mutex.release()


class RealtimeHttpClient(Document):
    def __init__(self, name, who = "realtime_http_client",session = {}, config = None):
        if config:
            set_config(config)
        super().__init__(name, who, io={"enter":http_enter,"delete":http_delete,"exit":http_exit,"flush":http_flush,"sync":http_sync}, client=True, session=session)


from setuptools import setup


setup(
    name='pyappi',
    version='1.1.9',
    packages=['pyappi', 
              'tests',
              'pyappi.client',
              'pyappi.logs',
              'pyappi.block',
              'pyappi.document',
              'pyappi.encoding',
              'pyappi.stats',
              'pyappi.sync',
              'pyappi.user',
              'pyappi.util',
              'pyappi.service',
              'pyappi.events'],
    license='Copyright (c) All Rights Reserved',
    description="Native Python Appi implementation. Single threaded, single node, no plug-ins.",
    long_description=open('README.md').read(),
    entry_points={
        'console_scripts': ['pyappi=pyappi.server_cli:main',
                            'pyappi-client=pyappi.client_cli:main'
                            ],
    },
    install_requires=[
        "wheel",
        "uvicorn",
        "twine",
        "fastapi",
        "pathvalidate",
        "twine",
        "colorama",
        "httpx",
        "click",
        "websockets"
    ],
)



from pyappi.api_base import app
from pyappi.block.endpoints import *
from pyappi.user.session import verify_session  
from pyappi.document.access import document_with_read_access, document_with_write_access, document_with_owner_access
from fastapi import Response
from pyappi.document.merge import server_merge
from pyappi.document.enumeration import  is_new_document, lookup_document_id
from pyappi.stats.stats import StatsRecord
from pyappi.document.delta import changes_since_dict


@app.get("/document/{id}")
@verify_session
async def get_document(id, request: Request):
    if id.startswith("!"):
        id = lookup_document_id(id[1:].split(".")[0])

    result = document_with_read_access(id, request.state.session)

    return result if result else Response(status_code=404)


@app.get("/document/delta/{id}/{tsx}")
@verify_session
async def get_delta(id, tsx, request: Request):
    if id.startswith("!"):
        id = lookup_document_id(id[1:].split(".")[0])

    result = document_with_read_access(id, request.state.session)

    return changes_since_dict(result,int(tsx)) if result else Response(status_code=404)


@app.delete("/document/{id}")
@verify_session
async def delete_document(id, request: Request):
    def handler(document):

        stats = StatsRecord(id)
        
        stats.delete()
        document.delete()

        return Response(status_code=204)

    return document_with_write_access(id, request.state.session, handler)


@app.put("/document/{id}")
@verify_session
async def update_document(id, request: Request):
    write = await request.json()

    def handler(document):
        if document._cmt and write["_bmt"] != -1 and document._cmt != write["_bmt"]:
            return Response(status_code=409)
        
        del write["_bmt"]
            
        server_merge(document, write)

        return Response(status_code=202)

    return document_with_write_access(id, request.state.session, handler)


@app.post("/document/{id}")
async def create_document(id, request: Request):
    write = await request.json()

    def handler(document):
        if not is_new_document(document):
            return Response(status_code=409)

        if document._cmt and write["_bmt"] != -1 and document._cmt != write["_bmt"]:
            return Response(status_code=409)
        
        del write["_bmt"]
        
        # prepare_new_document(document,id)
            
        server_merge(document,write)

        return Response(status_code=202)

    return document_with_write_access(id, request.state.session, handler)



from threading import RLock

import os
import json


event_config = {
    "root": "appidb/events",
}


if not os.path.exists(event_config["root"]):
    os.makedirs(event_config["root"],exist_ok=True)

    try:
        with open(event_config["root"] + "/events.json", "w") as doc:
            doc.write(json.dumps({
                "start": 0,
                "events": {}
            } , indent=4))
    except Exception as e:
        pass


event_lock = RLock()


def publish_event(event):
    MAX_EVENTS = 10000
    overflow = False
    with event_lock:
        head_file = event_config["root"] + "/events.json"
        with open(head_file, "r") as doc:
            db = json.load(doc)
            events = db["events"]
            current = len(events)

            current += 1

            events[str(current)] = event

            if current > MAX_EVENTS:
                overflow = True
                offset = db["start"] + MAX_EVENTS
                archive_file = f'{event_config["root"]}/events{db["start"]}.json'

        if overflow:
            with open(archive_file, "w") as doc:
                doc.write(json.dumps(db, indent=4))

            with open(head_file, "w") as doc:
                doc.write(json.dumps({
                    "start": offset,
                    "events":{}
                } , indent=4))
        else:
            with open(head_file, "w") as doc:
                doc.write(json.dumps(db , indent=4))

def read_cursor():
    head_file = event_config["root"] + "/events.json"
    with open(head_file, "r") as doc:
        db = json.load(doc)

        events = db["events"]
        current = len(events)

        return {"cursor": db["start"] + current + 1}


def read_events(start):
    MAX_EVENTS = 10000
    result = {"archive": False, "events": {}}
    events = result["events"]

    def copy_events(db, start):
        event = db["events"].get(str(start), None)
        while event:
            events[start] = event
            start += 1
            event = db["events"].get(str(start), None)

    with event_lock:
        head_file = event_config["root"] + "/events.json"
        with open(head_file, "r") as doc:
            db = json.load(doc)

            if db["start"] > start:
                result["archive"] = True
            else:
                copy_events(db, start)

        if result["archive"]:
            archive_file = f'{event_config["root"]}/events{start%MAX_EVENTS}.json'
            with open(archive_file, "r") as doc:
                db = json.load(doc)

                copy_events(db, start)

    return result


from pyappi.api_base import app
from pyappi.block import read1, write1, is1, read2, write2, is2

from fastapi import Response, Depends, Request

async def get_body(request: Request):
    return await request.body()

@app.get("/block/{type}/{id}")
async def get_block(type, id):
    block = read1(type,id)

    return Response(status_code=404) if not block else Response(content=block, media_type='application/octet-stream')

@app.get("/block2/{type}/{id}")
async def get_block2(type, id):
    block = read2(type,id)

    return Response(status_code=404) if not block else Response(content=block, media_type='application/octet-stream')

@app.get("/is/{type}/{id}")
async def get_is(type, id):
    return is1(type,id)

@app.get("/is2/{type}/{id}")
async def get_is2(type, id):
    return is2(type,id)

@app.post("/block/{type}")
async def post_block(type, body: bytes = Depends(get_body)):
    
    id = write1(type, body)
    return {"id":id, "result":0}

@app.post("/block2/{type}/{id}")
async def post_block(type, body: bytes = Depends(get_body)):
    
    write2(type, id, body)
    return {"id":id, "result":0}


from pyappi.document.transaction import Transaction

def server_merge(document, write):
    for k, _ in write.items():
        o = document.get(k,None)
        if isinstance(write[k], dict) and (not isinstance(document,Transaction) or (isinstance(o, Transaction) or isinstance(o, dict))): 
            server_merge(document[k], write[k])
        else:
            document[k] = write[k]
    


from pyappi.document.lock import global_appi_mutex
from pyappi.document import Document


volatile_documents = {}


def volatile_enter(self):
    global_appi_mutex.acquire()

    self.__dict__['__lock'] = True
    self.__dict__['__document'] = volatile_documents.get(self.__dict__['name'],{})
    self.__dict__['__tsx'] = self.__dict__['__document'].get("_cmt",0) + 1

    return self

def volatile_sync(self):
    if self.__dict__['__lock'] != True:
        raise "Document not locked"
    
    if len(self.__dict__['__delta']):
        volatile_flush(self)

    # Optional: must readback only if multi threaded access is implemented


def volatile_delete(self):
    try:
        del volatile_documents[self.__dict__['name']]
        return True
    except Exception as _:
        return False


def volatile_flush(self):
    if not self.__dict__['__mutated'] or self.__dict__["__read_only"]:
        return
    
    volatile_documents[self.__dict__['name']] = self.__dict__['__document']
    self.__dict__['__delta'] = {}


def volatile_exit(self, type, value, traceback):
    volatile_flush(self)

    self.__dict__['__lock'] = False
    global_appi_mutex.release()


class VolatileDocument(Document):
    def __init__(self, name, who = "volatile_document",session = {}):
        super().__init__(name, who, io={"enter":volatile_enter,"delete":volatile_delete,"exit":volatile_exit,"flush":volatile_flush,"sync":volatile_sync}, session=session)